;;
;; simple-md5 - Scheme wrapper for Colin Plumb's Public Domain md5 implementation, without dependencies
;;
;; All code in this egg is in the Public Domain
(module simple-md5 (string->md5sum file-md5sum)

(import scheme (chicken base) (chicken file)
        (chicken foreign) (chicken file posix) (chicken fixnum)
        memory-mapped-files srfi-4)

(foreign-declare "#include \"md5-base.c\"")

(define digest-length (foreign-value "MD5_DIGEST_SIZE" unsigned-int))
(define context-size (foreign-value "sizeof(struct MD5Context)" unsigned-int))

(define init (foreign-lambda void MD5Init bytevector))
(define update (foreign-lambda void MD5Update nonnull-u8vector nonnull-c-string unsigned-int))
(define raw-update (foreign-lambda void MD5Update nonnull-u8vector nonnull-c-pointer unsigned-int))
(define final (foreign-lambda void MD5Final nonnull-u8vector nonnull-u8vector))

(define (char->hexdigits int)
  (if (fx>= int 16)
      (number->string int 16)
      (string-append "0" (number->string int 16))))

(define (string->md5sum str)
  (let ((str-size (string-length str))
        (ctxt (make-u8vector context-size))
        (digest (make-u8vector digest-length)))
    (init ctxt)
    (unless (zero? str-size)
      (update ctxt str str-size))
    (final ctxt digest)
    (apply string-append (map char->hexdigits (u8vector->list digest)))))

;; From simple-sha1
(cond-expand
  ((and windows (not cygwin))
   (begin
     (define read-into-buffer
       (foreign-lambda* bool ((int fd) (nonnull-u8vector buffer) (integer size))
	 "C_return(read(fd, buffer, size) == size);"))
     (define (mapped-pointer fname fd size k)
       (let ((buffer (make-u8vector size)))
	 (unless (read-into-buffer fd buffer size)
	   (error 'sha1sum "can not read file" fname))
	 (k (location buffer) void)))))
  (else
   (define (mapped-pointer fname fd size k)
     (let* ((mmap (map-file-to-memory #f size prot/read map/shared fd))
	    (ptr (memory-mapped-file-pointer mmap)))
       (k ptr (cut unmap-file-from-memory mmap))))))


(define (file-md5sum fname)
  (and (file-exists? fname)
       (not (directory? fname))
       (let* ((fd (file-open fname open/rdonly))
	      (fsize (file-size fd))
	      (ctxt (make-u8vector context-size))
	      (digest (make-u8vector digest-length)))
	 (init ctxt)
	 (unless (zero? fsize)
	   (mapped-pointer
	    fname fd fsize
	    (lambda (buffer cleanup)
	      (raw-update ctxt buffer fsize)
	      (cleanup))))
	 (final ctxt digest)
	 (file-close fd)
	 (apply string-append (map char->hexdigits (u8vector->list digest))))))
)
